#!/usr/bin/env ruby


require 'optparse'
require 'json'


# Parse the options
option_parser = OptionParser.new do |opts|
  opts.on("-e", "--elf ELF_FILE", "The elf file to patch") do |filename|
    @filename = filename
  end
  opts.on("-c", "--config-file CONFIG_FILE", "The json configuration file") do |configfile|
    @config_file = configfile
  end
end

begin
  option_parser.parse!
  raise OptionParser::InvalidOption, OptionParser::MissingArgument if @filename.nil?
  raise OptionParser::InvalidOption, OptionParser::MissingArgument if @config_file.nil?
rescue OptionParser::InvalidOption, OptionParser::MissingArgument
  puts option_parser.help
  exit
end

# Open and check the configuration file
begin
  puts "Reading the config file..."
  @config = JSON.parse(File.read(@config_file))
  # Check the config file
  raise StandardError, 'aid_size field not set' if @config["aid_size"].nil? or
                                                not @config["aid_size"].is_a? Integer
  raise StandardError, 'blob_structure_name field not set' if @config["blob_structure_name"].nil? or
                                                           not @config["blob_structure_name"].is_a? String
  raise StandardError, 'protect field not set' if @config["protect"].nil? or
                                               not @config["protect"].is_a? Array
  raise StandardError, 'protect field empty' if @config["protect"].empty?
  raise StandardError, 'attestator_number not set' if @config["attestator_number"].nil? or
                                                   not @config["attestator_number"].is_a? Integer
  if @config["protect"].select { |f| not f.is_a? String }.size > 0
    raise StandardError, 'Function names must be of type String'
  end
  raise StandardError, 'Duplicate function names' unless @config["protect"].uniq.size == @config["protect"].size
  raise StandardError, 'base_sym field not set' if @config["base_sym"].nil?

rescue StandardError => e
  puts "Unable to parse the config file"
  puts e.message
  exit
end


@label_starting_point = 0
@ra_data_structure_blob_size = @config["aid_size"] + 8
@ra_data_structure_blob_size += 2 + (2 + 4 + 8 + 4) * @config["protect"].uniq.size
@memory_areas_count = @config["protect"].uniq.size
raise StandardError, "There are too many memory areas to fit in 2 bytes" if @memory_areas_count > 0xffff



# Read the elf file and find the blob position
# Then search for all the symbols in the configuration file and get the memory offset and the size
begin
  require 'metasm'
  @patch_symbols = Hash.new

  e = Metasm::ELF.decode_file(@filename)

  puts "Check if the binary is of type ET_EXEC or ET_DYN..."
  raise StandardError, "File is not of type ET_EXEC or ET_DYN" if e.header.type != "EXEC" and e.header.type != "DYN"
  puts "Check Address Space Layout Randomization..."
  e.header.type == "EXEC" ? @aslr = false : @aslr = true
  puts "ASLR => #{@aslr}"


  puts "Check if the binary has a symtab..."
  sym_table = e.sections.find { |s| s.name == '.symtab' }
  raise StandardError, 'Looks like the binary is stripped!' if sym_table.nil?
  puts "Symbol table found"

  puts "Searching for #{@config["base_sym"]}..."
  base_sym = e.symbols.find_all { |s| s.name == @config["base_sym"] }[0]
  raise StandardError, 'Cannot find the symbol for the base address' if base_sym.nil?

  puts "Searching for #{@config["blob_structure_name"]}..."
  blob_sym =  e.symbols.find_all { |s| s.name == @config["blob_structure_name"] }[0]
  raise StandardError, 'Cannot find the symbol for the blob data structure' if blob_sym.nil?

  # Find the data segment and check the base_sym and the blob are in it
  data_seg = e.segments.find_all { |s|
    s.type == 'LOAD' and
    s.flags.include? "W" and
    s.flags.include? "R"
    s.vaddr < blob_sym.value and
    s.vaddr + s.memsz > blob_sym.value + blob_sym.size and
    s.vaddr < base_sym.value and
    s.vaddr + s.memsz > base_sym.value + base_sym.size  }[0]
  raise StandardError, 'Cannot find the segment where the blob and the base symbol are' if data_seg.nil?

  base_sym_offset = base_sym.value - data_seg.vaddr
  @base_sym_pos = data_seg.offset + base_sym_offset
  @base_sym_vaddr = base_sym.value
  puts "Base symbol found at offset 0x#{@base_sym_pos.to_s(16)}"

  blob_offset = blob_sym.value - data_seg.vaddr
  @blob_pos = data_seg.offset + blob_offset
  puts "Blob found at offset 0x#{@blob_pos.to_s(16)}"

  @text_seg = e.segments.find_all { |s|
    s.type == 'LOAD' and s.flags.include? "X" and s.flags.include? "R" }[0]
  raise StandardError, 'Cannot find the text segment' if @text_seg.nil?

  @config["protect"].uniq.each do |k|
    puts "Searching for symbol #{k}..."
    current_symbol =  e.symbols.find_all { |sym| sym.name == k}[0]
    raise StandardError, "Cannot find the symbol for the #{k} function" if blob_sym.nil?

    if current_symbol.value < @text_seg.vaddr or
        current_symbol.value + current_symbol.size > @text_seg.vaddr + @text_seg.memsz
      puts "Symbol #{k} is not in the text segment. Skipping..."
      next
    end
    
    @patch_symbols[current_symbol.name] = {
        :memory_offset => current_symbol.value - @text_seg.vaddr,
        :size => current_symbol.size
    }

    puts "Symbol #{k} found: #{@patch_symbols[current_symbol.name]}"
  end

rescue StandardError => e
  puts "Error parsing target file!"
  puts e.message
  exit
end


# Patch the blob with the information retrieved in the first part
begin
  puts "Patching the base symbol..."
  if @aslr
    puts "Aslr is enabled => Searching for the relocation entry..."
    require 'elftools'
    e = ELFTools::ELFFile.new(File.open(@filename))
    begin
      base_reloc = e.section_by_name('.rela.dyn').relocations.find { |s| s.header.r_offset == @base_sym_vaddr }
    rescue StandardError => e
      raise StandardError, 'Cannot find the relocation entry'
    end
    raise StandardError if base_reloc.nil?
    # puts base_reloc.header
    base_reloc.header.r_addend = @text_seg.offset
    # puts base_reloc.header
    e.save(@filename)
  else
    puts "Aslr is disabled => Patching the base symbol on disk..."
    File.write(@filename, [@text_seg.vaddr].pack('Q'), @base_sym_pos)
  end


  puts "Patching the blob..."
  # Skip the attestator ID and the attestator number
  offset = @blob_pos + @config["aid_size"]

  # Patch the attestator number  
  File.write(@filename, [@config["attestator_number"]].pack('Q'), offset)
  offset += 8

  # Patch the memory areas count

  File.write(@filename, [@memory_areas_count].pack('L'), offset)
  offset += 4

  # For each functions set the whole structure
  @patch_symbols.values.each do |ps|
    # Patch the label field
    File.write(@filename, [@label_starting_point+=1].pack('S'), offset)
    offset += 2

    #Patch the total_blocks field
    File.write(@filename, [1].pack('L'), offset)
    offset += 4

    # Patch the memory address
    File.write(@filename, [ps[:memory_offset]].pack('Q'), offset)    
    offset += 8

    # Patch the size
    File.write(@filename, [ps[:size]].pack('L'), offset)
    offset += 4
  end

  puts "Blob patched correctly!"

rescue StandardError => e
  puts "Error patching the target file!"
  puts e.message
  exit
end
