/*
 * 	Attestator IMPLEMENTATION
 */

#include "attestator.h"
#include "ra_blob.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "ra_do_hash.h"
#include "ra_data_preparation.h"
#include "ra_nonce_interpretation.h"
#include "ra_defines.h"
#include "ra_print_wrapping.h"

#define MAX_CONNECTION_RETRY 10


pthread_t ws_polling_thread_NAYjDD3l2s;
pthread_mutex_t websocket_mutex_NAYjDD3l2s;
pthread_mutex_t deinit_mutex_NAYjDD3l2s;

volatile int force_exit_NAYjDD3l2s;

pthread_cond_t init_finished_cond_NAYjDD3l2s;
pthread_mutex_t init_finished_mutex_NAYjDD3l2s;
volatile int init_finished_status_NAYjDD3l2s;

void *attestator_routine_NAYjDD3l2s() {

    pthread_mutex_lock(&deinit_mutex_NAYjDD3l2s);
    if (force_exit_NAYjDD3l2s == 1) return NULL;

    ra_fprintf(stdout, "(Attestator XXX) Starting attestation\n");
    ra_fflush(stdout);

    uint8_t *defined_AID;
    ra_get_AID_reference_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s, &defined_AID);

    uint64_t defined_attestator_number;
    ra_get_attestator_number_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s, &defined_attestator_number);



    /*** TIME_EVALUATION ***********************/
    float t0, t1;
    t0 = ((float) clock()) / CLOCKS_PER_SEC;
    /*******************************************/

    /* ------------------- ATTESTATION DATA PREPARATION ----------------------------- */
    uint64_t req_id = 0x0000000000000001;

    uint8_t nonce[8] = {0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8};

    uint32_t nonce_size = sizeof(nonce);
    ra_fprintf(stdout, "(Attestator %"PRIu64") Received nonce size: %"PRIu32"B\n", defined_attestator_number, nonce_size);
    ra_fflush(stdout);


    uint8_t *nonce_buffer = nonce;
    ra_fprintf(stdout, "(Attestator %"PRIu64") Received request id %"PRIu64" containing %"PRIu32"B nonce\n", defined_attestator_number, req_id,
               nonce_size);

    ra_set_current_nonce_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s, nonce_buffer, nonce_size);
    ra_fprintf(stdout, "(Attestator %"PRIu64") Nonce set in data structure\n", defined_attestator_number);
    ra_fflush(stdout);
    print_encoded_info_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);

    ra_fprintf(stdout, "(Attestator %"PRIu64") Launching data preparation\n", defined_attestator_number);
    if (ra_prepare_data_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s) != RA_SUCCESS) {
        ra_fprintf(stderr, "(Attestator %"PRIu64") Error while preparing data\n", defined_attestator_number);
        ra_fflush(stderr);
        return NULL;
    }

#ifdef HARD_DEBUG
    uint32_t dp_size = ra_get_prepared_data_size_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);
    uint8_t *dp = ra_get_prepared_data_reference_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);
    ra_fprintf(stdout, "(Attestator %"PRIu64") Prepared data size %"PRIu32"\n", defined_attestator_number,dp_size);
    for (int p = 0; p < dp_size; p++) {
        ra_fprintf(stdout, "%02X", (unsigned char) (*((int8_t*) (dp + p))));
    }
    ra_fprintf(stdout, "\n");
#endif

    uint8_t *packed_buffer;
    uint32_t packed_buffer_size;

    if (ra_do_hash_NAYjDD3l2s(
            ra_get_current_nonce_reference_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s),
            ra_get_current_nonce_size_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s),
            ra_get_prepared_data_reference_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s),
            ra_get_prepared_data_size_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s),
            defined_AID,
            defined_attestator_number,
            NULL,
            0,
            &packed_buffer,
            &packed_buffer_size
    ) != RA_SUCCESS) {
        ra_fprintf(stdout, "(Attestator %"PRIu64") Error hashing prepared data\n", defined_attestator_number);
        ra_fflush(stdout);
        return NULL;
    }

    ra_fprintf(stdout, "(Attestator %"PRIu64") Attestation data hashed, %"PRIu32"B buffer\n", defined_attestator_number, packed_buffer_size);

#ifdef HARD_DEBUG
    ra_fprintf(stdout, "(Attestator %"PRIu64") Hashed data size %"PRIu32"\n", defined_attestator_number,packed_buffer_size);
    for (int p = 0; p < packed_buffer_size; p++) {
        ra_fprintf(stdout, "%02X", (unsigned char) (*((int8_t*) (packed_buffer + p))));
    }
    ra_fprintf(stdout, "\n");
#endif
    /* ------------------- ATTESTATION DATA PREPARED ----------------------------- */




    /*
     * Destroy prepared data in order to keep it in memory
     * no longer than necessary
     */
    ra_reset_prepared_data_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);


    /* ------------------- ATTESTATION RESPONSE PREPARATION ----------------------------- */
    uint8_t *response_buffer = (uint8_t *) malloc((ATTESTATOR_RESP_HEADER_SIZE + packed_buffer_size) * sizeof(uint8_t));
    if (response_buffer == NULL) {
        ra_fprintf(stderr, "(Attestator %"PRIu64") Error during allocation\n", defined_attestator_number);
        ra_fflush(stderr);
        return NULL;
    }

    memcpy(response_buffer, defined_AID, AID_SIZE);
    ra_fprintf(stdout, "(Attestator %"PRIu64") Writing AID to response_buffer\n", defined_attestator_number);


    ra_fprintf(stdout, "(Attestator %"PRIu64") Writing attestator number to response_buffer\n", defined_attestator_number);
    memcpy(response_buffer + AID_SIZE, &defined_attestator_number, RESP_ATTESTATOR_NO_SIZE);


    ra_fprintf(stdout, "(Attestator %"PRIu64") Writing req_id to response_buffer\n", defined_attestator_number);
    memcpy(response_buffer + AID_SIZE + RESP_ATTESTATOR_NO_SIZE, &req_id, RESP_REQUEST_ID_SIZE);


    ra_fprintf(stdout, "(Attestator %"PRIu64") Writing packed_buffer_size to nonce_buffer\n", defined_attestator_number);
    memcpy(response_buffer + AID_SIZE + RESP_ATTESTATOR_NO_SIZE + RESP_REQUEST_ID_SIZE, &packed_buffer_size, RESP_HASH_LENGTH_SIZE);


    ra_fprintf(stdout, "(Attestator %"PRIu64") Writing packed_buffer to nonce_buffer\n", defined_attestator_number);
    memcpy(response_buffer + AID_SIZE + RESP_ATTESTATOR_NO_SIZE + RESP_REQUEST_ID_SIZE + RESP_HASH_LENGTH_SIZE, packed_buffer, packed_buffer_size);
    /* ------------------- ATTESTATION RESPONSE PREPARED ----------------------------- */



    free(packed_buffer);
    free(response_buffer);

    /*** TIME_EVALUATION ***********************/
    t1 = ((float) clock()) / CLOCKS_PER_SEC;
    ra_fprintf(stdout, "(Attestator %"PRIu64") Request processed in = %f s\n", defined_attestator_number, t1 - t0);
    ra_fflush(stdout);
    /*******************************************/
    pthread_mutex_unlock(&deinit_mutex_NAYjDD3l2s);
    return NULL;

}



void *ws_polling_routine_NAYjDD3l2s(void *useless) {

    uint64_t attestator_number;

    if (ra_get_attestator_number_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s, &attestator_number) != RA_SUCCESS) {
        ra_fprintf(stderr, "(Attestator init %"PRIu64") Error: not defined attestator number, attestator cannot start\n", attestator_number);
        ra_fflush(stderr);
        ra_destroy_table_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);
        return NULL;
    }


    ra_fprintf(stdout, "(Attestator polling %"PRIu64") Initialization of ACCL\n", attestator_number);
    ra_fflush(stdout);

    while(1) {
        attestator_routine_NAYjDD3l2s();
        sleep(3);
    }

}


int attestator_init_NAYjDD3l2s() {

    uint64_t attestator_number;
    base_address_NAYjDD3l2s = text_seg_offset;
    ra_fprintf(stdout, "(Attestator) An attestator is being started\n");
    ra_fflush(stdout);

    if ((remote_attestation_data_table_NAYjDD3l2s = ra_parse_binary_ads_NAYjDD3l2s()) == NULL) {
        ra_fprintf(stderr, "(Attestator init) Error while parsing ADS, attestator cannot start\n");
        ra_fflush(stderr);
        return 1;
    }

    ra_print_table_info_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);

    if (ra_get_attestator_number_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s, &attestator_number) != RA_SUCCESS) {
        ra_fprintf(stderr, "(Attestator init %"PRIu64") Error: not defined attestator number, attestator cannot start\n", attestator_number);
        ra_fflush(stderr);
        ra_destroy_table_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);
        return 1;
    }


    ra_fprintf(stderr, "(Attestator init %"PRIu64") Register attestator deinit function\n", attestator_number);
    ra_fflush(stderr);
    if (atexit(attestator_deinit_NAYjDD3l2s) != 0) {
        ra_fprintf(stderr, "(Attestator init %"PRIu64") Cannot register exit function, attestator cannot start\n", attestator_number);
        ra_fflush(stderr);
        return 1;
    }


    force_exit_NAYjDD3l2s = 0;
    init_finished_status_NAYjDD3l2s = 0;
    pthread_mutex_init(&websocket_mutex_NAYjDD3l2s, NULL);
    pthread_mutex_init(&init_finished_mutex_NAYjDD3l2s, NULL);
    pthread_mutex_init(&deinit_mutex_NAYjDD3l2s, NULL);
    pthread_cond_init(&init_finished_cond_NAYjDD3l2s, NULL);

    ra_fprintf(stdout, "(Attestator init %"PRIu64") Creating communication thread\n", attestator_number);
    ra_fflush(stdout);
    if (pthread_create(&ws_polling_thread_NAYjDD3l2s, NULL, ws_polling_routine_NAYjDD3l2s, NULL)) {
        ra_fprintf(stderr, "(Attestator init %"PRIu64") Error creating polling thread, attestator cannot start\n", attestator_number);
        ra_fflush(stderr);
        ra_destroy_table_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);
        return 1;
    }


    ra_fprintf(stdout, "(Attestator init %"PRIu64") Successfully started\n", attestator_number);
    ra_fflush(stdout);

    return 0;
}

void attestator_deinit_NAYjDD3l2s() {


    ra_fprintf(stdout,
               "(Attestator deinit XXX) Application closed, destroying attestation data\n");

    uint64_t attesator_number;
    ra_get_attestator_number_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s, &attesator_number);
    ra_fprintf(stdout,
               "(Attestator deinit %"PRIu64") Application closed, destroying attestation data\n",
               attesator_number);

    pthread_mutex_lock(&deinit_mutex_NAYjDD3l2s);

    force_exit_NAYjDD3l2s = 1;

    ra_fprintf(stdout, "(Attestator deinit %"PRIu64") RA thread termination induced\n", attesator_number);

    if (pthread_join(ws_polling_thread_NAYjDD3l2s, NULL) != 0) {
        ra_fprintf(stderr, "(Attestator deinit %"PRIu64")  Error while cancelling thread\n");
        exit(1);
    }


    ra_fprintf(stdout, "(Attestator deinit %"PRIu64") Destroying attestation data\n", attesator_number);
    ra_destroy_table_NAYjDD3l2s(remote_attestation_data_table_NAYjDD3l2s);
    
    pthread_mutex_unlock(&deinit_mutex_NAYjDD3l2s);
    pthread_mutex_destroy(&init_finished_mutex_NAYjDD3l2s);
    pthread_mutex_destroy(&deinit_mutex_NAYjDD3l2s);
    pthread_mutex_destroy(&websocket_mutex_NAYjDD3l2s);

    ra_fprintf(stdout, "(Attestator deinit %"PRIu64") Successfully destroyed attestation data\n", attesator_number);
    return;
}
